//
// N A M E : p a r t d u m p . C
//
// D E S C : A disk partition dump utility
//
// A U T H : Wataru Nishida, M.D., Ph.D.
//           wnishida@skyfree.org
//           http://www.skyfree.org
//
// M A K E : gcc -o partdump partdump.C xseek.o
//
// V E R S : 0.92
//
//           SCSI disks are not supported.
//           Does anyone know how to get disk geometries from /dev/sdx?
//
// D A T E : Jan. 7th, 2001
//
// L  O  G : v0.90 11/24/2000 Full code was stolen from dumpmbr.C.
//           v0.91 11/25/2000 Class "disk" was newly created.
//           v0.92 01/07/2001 Extended disk was supported.
//

#include <stdio.h>		// printf(), f...()
#include <fcntl.h>		// open()
#include <unistd.h>		// close()
#include <linux/unistd.h>	// loff_t
#include <string.h>		// rindex(), str...()
#include <ctype.h>		// isalpah()
#include "xseek.h"		// xseek(); llseek() by assembly

// Constant definitions

static const char Version[]      = "0.92";

static const int  SectorSize     = 512;		// Sector size.
static const int  MaxDevName     = 3;		// hd? | sd?
static const int  MaxPrimary     = 4;		// Number of pri. partitions.
static const int  MaxSecondary   = 32;		// Number of sec. partitions.
static const int  MaxPartitions  = (MaxPrimary + MaxSecondary);
static const int  MaxRecord      = MaxSecondary+1;

static const char PrefixIDE      = 'h';
static const char PrefixSCSI     = 's';
static const char ProcDevIDE[]   = "/proc/ide/";
static const char ProcDevSCSI[]  = "/proc/scsi/";
static const char ProcDevMedia[] = "/media";
static const char ProcDevDisk[]  = "disk";
static const char ProcDevSize[]  = "/capacity";
static const char ProcDevCHS[]   = "/geometry";

// "Extended partition" IDs

static const unsigned char ExtendedTags[] = { 5, 0x85, 0 };

//
// Disk record definitions
//
//   There are two kinds of special sectors on a hard disk.
//   One is Master Boot Record (MBR) and another is extended partition
//   record (EPR). Both MBR and EPR store information of four partition
//   tables, and so MBR and EPR have the same structure.
//   MBR is unique on the disk and it is holded in the 1st sector.
//   MBR also contains an Initial Program Loader (IPL) in it.
//   EPR is optional and it holds extended partition information.
//   There may be multiple EPRs on the disk, and they hold secondary
//   partition tables.
//
//   I feel "MBR and EPR" are quite confusing names, and propose to
//   designate them as Original Disk Record (ODR) and Extended Disk Record
//   (EDR), respectively.
//
//   ODR holds 0-4 primary partition tables and 0-1 extended disk table
//   (EDT). EDT points EDR and its LBA address is counted from the beginning
//   of the EDR. EDR contains 1 secondary partition table and 0-1 EDT. 
//   Thus, secondary partition tables are stored in the "linked-list".
//
//   I encapsulated the disk record structure and its related functions into
//   C++ "class", instead of using old-styled "struct".
//   Please confirm the result of `ldd partdump`. As you can see, this code
//   does NOT require additional C++ library.
//
//   Wataru Nishida
//						Jan. 6, 2001

// 32bits LBA (it would be too small soon...)

typedef unsigned long lba32;	// NOTICE: LBA starts from ZERO!

//
// class
//
//   c h s
//
//   --- holds CHS information
//
//						Nov. 25, 2000

class chs {
public:
  int c, h, s;
 };

//
// class
//
//   c h s B I O S
//
//   --- holds encoded CHS information for BIOS
//
//						Nov. 25, 2000

class chsBIOS {
private:
  unsigned char data[ 3 ];
public:
  int getC(void) { return( ((data[ 1 ] & 0xC0) << 2) + data[ 2 ] ); }
  int getH(void) { return( data[ 0 ] ); }
  int getS(void) { return( data[ 1 ] & 0x3F); }
 };

//
// class
//
//   p a r t i t i o n
//
//   --- holds processed partition table information
//
//						Nov. 25, 2000

class partition {
public:
  int     boot,			// Boot flag.
          type;			// Filesystem type.
  lba32   size;			// Partition size in sectors.
  lba32   startLBA, endLBA;	// LBA geometry.
  chs     startCHS, endCHS;	// CHS geometry.
 };

//
// class
//
//   p a r t i t i o n T B L
//
//   --- holds and manipulate "raw" partition table data
//
//						Nov. 25, 2000

class partitionTBL {
private:
  unsigned char data[ 16 ];
public:
  int getBoot(void) { return( data[ 0 ] ); }
  int getType(void) { return( data[ 4 ] ); }
  chsBIOS* getStartCHS(void) { return( (chsBIOS*) &data[ 1 ] ); }
  chsBIOS* getEndCHS(void) { return( (chsBIOS*) &data[ 5 ] ); }
  lba32* getStartLBA(void) { return( (lba32*) &data[ 8 ] ); }
  lba32* getSize(void) { return( (lba32*) &data[ 12 ] ); }
 };

//
// class
//
//   d i s k R e c
//
//   --- holds disk record information (ODR and EDR)
//
//						Jan. 6, 2001

class diskRec {
public:
  unsigned char buffer[ SectorSize ];	// Sector buffer.
  lba32 offset;				// Offset address in sectors.
  chs   pos;				// Position in CHS.
  int   extdisk;			// Extended disk table index.

public:
  partitionTBL* getPartPtr(int idx) \
	{ return( (partitionTBL*) &buffer[ 0x1be + 16 * (idx-1) ]); }
  unsigned int getSign(void) \
	{ return( (buffer[ 0x1ff ] << 8) + buffer[ 0x1fe ]); }
 };

//
// class
//
//   d i s k
//
//   --- handle hard disk information
//
//						Nov. 25, 2000
//

class disk {
private:
  char      msg[ 128 ];			// Error message.
public:
  char      name[ MaxDevName + 1 ];	// Disk name.
  lba32     capacity;			// Size in sectors.
  chs       realCHS, biosCHS;		// Geometries.
  int       diskcnt;			// How many disk records are there.
  diskRec   drs[ MaxSecondary + 1 ];	// Disk records holder.
  int       partcnt;			// How many partitions are there.
  partition parts[ MaxPartitions ];	// Partitions holder.

private:
  int   readProc(void);			// Read /proc/ information.
  int   readOriginal(void);		// Read original disk record.
  bool  isExtended(int);		// Is this an extended disk table?
  int   readExtended(int);		// Read extended disk record.
public:
  int   readRecord(const char* dev);	// Read all of disk records.
  void  dumpRecord(int);		// Dump disk record contents.
  void  showInfo(void);			// Show disk information.
  char* message(void) { return(msg); }	// Tell error message.
 };

//
// class disk : private
//
//   r e a d P r o c
//
//   --- read /proc/ files and get physical drive information
//   +++ Currently, SCSI drive is not supported.
//
//						Nov. 25, 2000

int disk::readProc(void)
 {
  const char* procDev;
  char  path[ 64 ], temp[ 64 ], buf[ 64 ], tag[ 32 ];
  FILE* proc;
  int   ret,
        found, c, h, s;
  lba32 size;

  // IDE or SCSI?
  if (*name ==  PrefixIDE)		// IDE drive? 
    procDev = ProcDevIDE;
//--- FIXME ---
//else if (*name == PrefixSCSI)		// SCSI drive? 
//  return(0);
  else {
    strcpy(msg, "Unsupported device");
    return(1);
   } 
  strcpy(path, procDev);
  strcat(path, name);

  // Check media type

  strcpy(temp, path);
  strcat(temp, ProcDevMedia);
  proc = fopen(temp, "r");
  if (! proc) {
    strcpy(msg, "Failed to open ");
    strcat(msg, temp);
    return(1);
   }
  ret = fscanf(proc, "%32s", buf);	// Read media type.
  fclose(proc);
  if (ret != 1) {
    strcpy(msg, "Failed to read ");
    strcat(msg, temp);
    return(1);
   }
  if (strcmp(buf, ProcDevDisk) != 0) {	// Disk?
    strcpy(msg, "Media type is ");
    strcat(msg, buf);
    return(1);
   }

  // Get disk capacity

  strcpy(temp, path);
  strcat(temp, ProcDevSize);
  proc = fopen(temp, "r");
  if (! proc) {
    strcpy(msg, "Failed to open ");
    strcat(msg, temp);
    return(1);
   }
  ret = fscanf(proc, "%lu", &size);	// Read disk size in sectors.
  fclose(proc);
  if (ret != 1) {
    strcpy(msg, "Failed to read ");
    strcat(msg, temp);
    return(1);
   }
  capacity = size;

  // Get physical and logical geometries

  strcpy(temp, path);
  strcat(temp, ProcDevCHS);
  proc = fopen(temp, "r");
  if (! proc) {
    strcpy(msg, "Failed to open ");
    strcat(msg, temp);
    return(1);
   }
  while (fscanf(proc, "%32s %d/%d/%d", tag, &c, &h, &s) == 4) {
    if (strcmp(tag, "physical") == 0) {
      realCHS.c = c; realCHS.h = h; realCHS.s = s;
     }
    if (strcmp(tag, "logical") == 0) {
      biosCHS.c = c; biosCHS.h = h; biosCHS.s = s;
      found = 1;
     }
   }
  fclose(proc);
  if (! found) {
    strcpy(msg, "Failed to read ");
    strcat(msg, temp);
    return (1);
   }
  return(0);
 }

//
// class disk : private
//
//   r e a d O r i g i n a l
//
//   --- read Original Disk Record (MBR) and extract information
//
//						Nov. 25, 2000

int disk::readOriginal(void)
 {
  char path[ 64 ];
  int  fd, ret, i;
  long pos;
  partitionTBL* pt;

  // Reset counters

  diskcnt = 0;				// Reset disk record counter.
  drs[ diskcnt ].offset = 0UL;		// Offset address of ODR is zero.
  drs[ diskcnt ].extdisk = 0;		// Reset extended disk index.

  drs[ diskcnt ].pos.c = 0;		// ODR position is 0/0/1.
  drs[ diskcnt ].pos.h = 0;
  drs[ diskcnt ].pos.s = 1;

  // Read the Original Disk Record

  strcpy(path, "/dev/");
  strcat(path, name);
  fd = open(path, O_RDONLY);
  if (fd < 0) {
    strcpy(msg, "Can't open the device");
    return(1);
   }
  pos = lseek(fd, 0L, SEEK_SET);
  if (pos != 0L) {
    close(fd);
    strcpy(msg, "Seek error");
    return(1);
   }
  ret = read(fd, &drs[ diskcnt ].buffer, SectorSize);
  close(fd);
  if (ret != SectorSize) {
    strcpy(msg, "ODR read error");
    return(1);
   }

  // Scan the disk record

  for (i=0; i < MaxPrimary; i++) {	// Examine all table entries.
    pt = drs[diskcnt].getPartPtr(i+1);	// Points the table.
    if (isExtended(pt->getType()))	// Is this an extended disk table?
      drs[ diskcnt ].extdisk = i + 1;	// Then save its index number.
   }
  diskcnt++;				// Increment disk record counter.
  return(0);
 }

//
// class disk : private
//
//   i s E x t e n d e d
//
//   --- check this is an extended disk table or not.
//
//   [ int ] partition type number
//
//   [ return ] boolean
//						Dec. 1, 2000

bool disk::isExtended(int type)
 {
  int i = 0;
 
  while (ExtendedTags[ i ]) {
    if (ExtendedTags[ i ] == type) return( true );
    else i++;
   }
  return( false );
 }

//
// class disk : private
//
//   r e a d E x t e n d e d
//
//   --- read information of extended disk record (EDR)
//
//   [ int ] Extended disk table index number (starts from one)
//
//						Jan. 7, 2001

int disk::readExtended(int idx)
 {
  char path[ 64 ];
  unsigned int fd;
  int  ret, i;
  loff_t pos, res;
  lba32 dr_offset,		// Disk record offset address.
        sec_cyl, sec_track;	// Disk geometry.
  partitionTBL* pt;

  // Read the Extended Disk Record

  strcpy(path, "/dev/");
  strcat(path, name);
  fd = open(path, O_RDONLY);
  if (fd < 0) {
    strcpy(msg, "Can't open the device");
    return(1);
   }

  pt = drs[ diskcnt-1 ].getPartPtr(idx);	// Get table address.

// I've ever found a bug in LBA management of FDISK.
// "CHS tells the truth..."

#if 0
  dr_offset = *(pt->getStartLBA());		// Read starting LBA address.
  dr_offset += drs[ diskcnt-1 ].offset;		// Add disk record offset.
#else
  sec_cyl   = biosCHS.h * biosCHS.s;		// Sectors / cylinder.
  sec_track = biosCHS.s;			// Sectors / track.
  dr_offset =  (lba32) ((pt->getStartCHS())->getC()) * sec_cyl;
  dr_offset += (lba32) ((pt->getStartCHS())->getH()) * sec_track;
  dr_offset += (lba32) ((pt->getStartCHS())->getS()) - 1UL;
#endif

  drs[ diskcnt ].pos.c = (pt->getStartCHS())->getC();
  drs[ diskcnt ].pos.h = (pt->getStartCHS())->getH();
  drs[ diskcnt ].pos.s = (pt->getStartCHS())->getS();

  pos = ((loff_t) dr_offset) * SectorSize;	// Calculate byte address.
  ret = xseek(fd, &pos, &res, SEEK_SET);
  if (res != pos) {
    close(fd);
    strcpy(msg, "Seek error");
    return(1);
   }
  ret = read(fd, &drs[ diskcnt ].buffer, SectorSize);
  close(fd);
  if (ret != SectorSize) {
    strcpy(msg, "Extended disk record read error");
    return(1);
   }
  drs[ diskcnt ].offset = dr_offset;	// Save offset address.
  drs[ diskcnt ].extdisk = 0;		// Reset extended disk index.

  // Scan the disk record

  for (i=0; i < MaxPrimary; i++) {	// Examine all table entries.
    pt = drs[diskcnt].getPartPtr(i+1);	// Points the table.
    if (isExtended(pt->getType()))	// Is this an extended disk table?
      drs[ diskcnt ].extdisk = i + 1;	// Then save its index number.
   }
  diskcnt++;				// Increment disk record counter.
  return(0);
 }

//
// class disk : public
//
//   r e a d R e c o r d
//
//   --- read disk record information according to "linked-list"
//
//						Jan. 7, 2001

int disk::readRecord(const char* dev)
 {
  const char* ptr;
  int ret;

  // Check device name
  ptr = dev;
  while (*ptr) {
    if (! isalpha(*ptr)) {
      strcpy(msg, "Illegal device name");
      return(1);
     }
    ptr++;
   }
  if (strlen(dev) != (size_t) MaxDevName) {
    strcpy(msg, "Illegal device name");
    return(1);
   }
  strcpy(name, dev);			// Save device name.

  ret = this->readProc();		// Read /proc/ information.
  if (ret) return(ret);

  ret = this->readOriginal();		// Read Original Disk Record.
  if (ret) return(ret);

  while ( true ) {			// Follow the "linked-list".
    if (diskcnt > MaxRecord) {		// Disk record overflow?
      strcpy(msg, "Too many disk records");
      return(1);
     }
    if (drs[ diskcnt-1 ].extdisk) {	// There is a extended disk?
      ret = this->readExtended(drs[ diskcnt-1 ].extdisk);
      if (ret) return(ret);
     }
    else break;    			// End of list.
   }

  return(0);
 }

//
// class disk : public
//
//   d u m p R e c o r d
//
//   --- dump disk records
//
//   [ int ] Disk record index number (starts from ZERO)
//
//                                              Jan. 7, 2001

void disk::dumpRecord(int idx)
 {
  partitionTBL* pt;
  int sign, c, h, s;

  sign = drs[ idx ].getSign();		// Read signature.
  if (! idx)
    printf("Original Disk Recrod ");
  else
    printf("Extended Disk Recrod ");
  printf("(Signature %04X) ", sign);
  printf("@ LBA %ld / ", drs[ idx ].offset);
  printf("CHS %d/%d/%d\n", drs[idx].pos.c, drs[idx].pos.h, drs[idx].pos.s);

  for (int i=0; i < MaxPrimary; i++) {	// Scan all entries.
    pt = drs[ idx ].getPartPtr( i+1 );	// Points each table entry.
    if (this->isExtended(pt->getType()))
      printf("  * ");
    else
      printf("    ");

    printf("%d) Boot %02X : ", i+1, pt->getBoot());
    printf("Type %02X : ", pt->getType());
    printf("%11ld sectors : ", *(pt->getSize()));
    printf("%ld MB \n", *(pt->getSize())/2000UL);

    printf("       CHS ");
    c = (pt->getStartCHS())->getC();
    h = (pt->getStartCHS())->getH();
    s = (pt->getStartCHS())->getS();
    printf("%4d/%3d/%2d --> ", c, h, s);
    c = (pt->getEndCHS())->getC();
    h = (pt->getEndCHS())->getH();
    s = (pt->getEndCHS())->getS();
    printf("%4d/%3d/%2d\n", c, h, s);

    printf("       LBA ");
    printf("%11ld --> ", *(pt->getStartLBA()));
    printf("%11ld\n", *(pt->getStartLBA()) + *(pt->getSize()) - 1UL);
   }
  printf("\n");
 }

//
// class disk : public
//
//   s h o w I n f o
//
//   --- show disk information
//
//						Nov. 25, 2000

void disk::showInfo(void)
 {
  printf("[ /dev/%s ]\n\n", name);

  // Currently only IDE drive specs are available through /proc/

  if (*name ==  PrefixIDE) {		// IDE drive? 
    printf("  Total %ld sectors : ", capacity);
    printf("%ld MB (1KB = 1024)\n", capacity/2000);
    printf("  Real CHS %d/%d/%d : ", realCHS.c, realCHS.h, realCHS.s);
    printf("BIOS CHS %d/%d/%d\n\n", biosCHS.c, biosCHS.h, biosCHS.s);
   }
 }

//
// u s a g e
//
//   --- show my usage
//
//						Nov. 24, 2000

static void usage(void)
 {
  printf("-- partdump v%s -- ", Version);
  printf("Partition dump utility.\n");
  printf("Usage:  partdump <device>\n");
  exit(1);
 }

//
// e r r o r
//
//   --- in case of a fatal error, report the situation and abort
//
//						Nov. 24, 2000

static void error(const char* msg)
 {
  printf("prdump: %s\n", msg);
  exit(1);
 }

//
// m a i n
//
//   --- Welcome to partdump!
//
//						Jan. 7, 2001

int main(int argc, char** argv)
 {
  disk drive;
  char *dev;
  int ret;
  
  // Check argument

  if (argc != 2) usage();
  dev = rindex(argv[1], '/');
  dev++;		// Skip /dev/.

  // Read into all disk records

  ret = drive.readRecord(dev);
  if (ret) error(drive.message());

  // Output disk information in case of IDE drive

  drive.showInfo();

  // Output disk records

  for (int i = 0; i < drive.diskcnt; i++)
    drive.dumpRecord( i );

  exit(0);
 }
